@@ -7,44 +7,56 @@ import shutil
 from pathlib import Path
 from tempfile import mkstemp
 from types import TracebackType
-from typing import Any, Iterable, Mapping, Sequence, TypeVar
+from typing import TYPE_CHECKING, Any, Iterable, Mapping, Sequence, TypeVar
 
 import pytest
 from sqlalchemy import text
-from sqlalchemy.engine import Connection, Engine, create_engine
+from sqlalchemy.engine import Connection, Engine, Row, create_engine
 from sqlalchemy.orm import sessionmaker
 from sqlalchemy.sql import insert, select
 from sqlalchemy.sql.schema import MetaData, Table
 
 from .builder import DatabaseBuilder
 from .session import Session
-from .types import RowType
+
+if TYPE_CHECKING:
+    from sqlalchemy.sql._typing import _DMLColumnArgument
 
 _S = TypeVar("_S", bound="DBFixture")
+_TP = TypeVar("_TP", bound="tuple[Any, ...]")
 
 _MEMORY_DB_URL = "sqlite:///:memory:"
 
+# Placeholder for any NOT NULL value.
+NOT_NULL = object()
 
-def assert_row_equals(
-    row: Mapping[str, Any], expected_values: Mapping[str, Any]
-) -> None:
+
+def assert_row_equals(row: Row[Any], expected_values: Mapping[str, Any]) -> None:
     """Assert that a row contains expected values.
 
     row is a mapping as returned from execute_sql_one_row() or
     fetch_only_row(). expected_values is a column name -> expected value
     mapping. The row may contain additional columns that are not listed
     in expected_values. Those are ignored.
+
+    The expected value may be NOT_NULL, in which case any value that is not
+    None (i.e. NULL in the database) is considered valid.
     """
 
     for column_name, expected in expected_values.items():
-        column_value = row[column_name]
-        assert (
-            column_value == expected
-        ), f"column '{column_name}': {expected!r} != {column_value!r}"
+        column_value = row._mapping[column_name]
+        if expected is NOT_NULL:
+            assert column_value is not None, (
+                f"column '{column_name}': expected NOT NULL, but was NULL"
+            )
+        else:
+            assert column_value == expected, (
+                f"column '{column_name}': {expected!r} != {column_value!r}"
+            )
 
 
 def assert_one_row_equals(
-    rows: Iterable[Mapping[str, Any]], expected_values: Mapping[str, Any]
+    rows: Iterable[Row[Any]], expected_values: Mapping[str, Any]
 ) -> None:
     """Assert that one of a list of rows contains the expected values.
 
@@ -54,7 +66,7 @@ def assert_one_row_equals(
     in expected_values. Those are ignored.
     """
 
-    def check_row(row: Mapping[str, Any]) -> bool:
+    def check_row(row: Row[Any]) -> bool:
         try:
             assert_row_equals(row, expected_values)
         except AssertionError:
@@ -121,8 +133,7 @@ class DBFixture:
             self._tmp_db = _copy_database(self.db_path)
         self.engine = create_engine(self.db_url)
         self._connection = self.engine.connect()
-        self._connection.execute(text("PRAGMA foreign_keys=ON"))
-        self.__metadata__.bind = self.engine
+        self.execute_sql("PRAGMA foreign_keys=ON")
         self._session = Session(sessionmaker(bind=self.engine)).__enter__()
 
         if self.sql_path is not None:
@@ -139,12 +150,14 @@ class DBFixture:
         exc_val: BaseException | None,
         exc_tb: TracebackType | None,
     ) -> None:
+        assert self.engine is not None
         assert self._session is not None
         assert self._connection is not None
         self._session.__exit__(exc_type, exc_val, exc_tb)
         self._session = None
         self._connection.close()
         self._connection = None
+        self.engine.dispose()
         self.engine = None
         if self._tmp_db is not None:
             os.remove(self._tmp_db)
@@ -175,9 +188,7 @@ class DBFixture:
         with self.connection.begin():
             self._db_builder.require(*features)
 
-    def execute_sql(
-        self, query: str, args: Mapping[str, Any] | None = None
-    ) -> None:
+    def execute_sql(self, query: str, args: Mapping[str, Any] | None = None) -> None:
         """Execute a SQL query."""
         with self.connection.begin():
             if args is None:
@@ -188,7 +199,7 @@ class DBFixture:
 
     def select_sql(
         self, query: str, args: Mapping[str, Any] | None = None
-    ) -> list[RowType]:
+    ) -> Sequence[Row[Any]]:
         """Execute a SQL SELECT and return all rows."""
         with self.connection.begin():
             if args is None:
@@ -202,7 +213,7 @@ class DBFixture:
 
     def select_sql_one_row(
         self, query: str, args: Mapping[str, Any] | None = None
-    ) -> RowType:
+    ) -> Row[Any]:
         """Execute a SQL SELECT and return one row.
 
         Raise an AssertionError if the result has zero or more than one row.
@@ -211,41 +222,45 @@ class DBFixture:
         assert len(rows) == 1, f"got {len(rows)} rows, expected 1"
         return rows[0]
 
-    def select_all_rows(self, table_name: str) -> list[RowType]:
+    def select_all_rows(self, table_name: str) -> Sequence[Row[Any]]:
         """Return all rows from a table."""
-        table = Table(table_name, self.__metadata__, autoload=True)
-        return self.select_sql(select([table]))
+        table = Table(table_name, self.__metadata__, autoload_with=self.engine)
+        with self.connection.begin():
+            res = self.connection.execute(select(table))
+            return res.fetchall()
 
-    def select_only_row(self, table_name: str) -> RowType:
+    def select_only_row(self, table_name: str) -> Row[Any]:
         """Return the only row from a table.
 
         Raise an AssertionError if the table has zero or more than one row."""
         rows = self.select_all_rows(table_name)
         assert len(rows) == 1, (
-            f"expected exactly one row in table '{table_name}', "
-            f"got {len(rows)}"
+            f"expected exactly one row in table '{table_name}', got {len(rows)}"
         )
         return rows[0]
 
-    def insert(self, table_name: str, values: Mapping[str, Any]) -> None:
+    def insert(
+        self,
+        table_name: str,
+        values: dict[_DMLColumnArgument, Any] | Sequence[Any],
+    ) -> None:
         """Insert one row into a table using a mapping of values."""
-        table = Table(table_name, self.__metadata__, autoload=True)
-        self.connection.execute(insert(table, values))
+        table = Table(table_name, self.__metadata__, autoload_with=self.engine)
+        with self.connection.begin():
+            self.connection.execute(insert(table).values(values))
 
     def assert_table_is_empty(self, table_name: str) -> None:
         """Assert that a table has no rows."""
         rows = self.select_all_rows(table_name)
         assert len(rows) == 0, (
-            f"table {table_name} contains {len(rows)} rows, "
-            "expected it to be empty"
+            f"table {table_name} contains {len(rows)} rows, expected it to be empty"
         )
 
     def assert_row_count(self, table_name: str, expected_rows: int) -> None:
         """Assert that a table has a certain amount of rows."""
         rows = self.select_all_rows(table_name)
         assert len(rows) == expected_rows, (
-            f"table {table_name} contains {len(rows)} rows, "
-            f"expected {expected_rows}"
+            f"table {table_name} contains {len(rows)} rows, expected {expected_rows}"
         )
 
     def assert_only_row_equals(
@@ -287,11 +302,13 @@ class DBFixture:
         fetched_rows = self.select_all_rows(table_name)
         if not fetched_rows and not expected_rows:
             return
-        assert len(fetched_rows) == len(expected_rows)
+        assert len(fetched_rows) == len(expected_rows), (
+            f"expected {len(expected_rows)} rows, got {len(fetched_rows)}"
+        )
 
         def find_one(
-            rs: list[RowType], expected: Mapping[str, Any]
-        ) -> list[RowType]:
+            rs: Sequence[Row[_TP]], expected: Mapping[str, Any]
+        ) -> list[Row[_TP]]:
             __tracebackhide__ = True
             for i, tr in enumerate(rs):
                 try:
@@ -299,7 +316,7 @@ class DBFixture:
                 except AssertionError:
                     pass
                 else:
-                    return rs[:i] + rs[i + 1 :]
+                    return [*rs[:i], *rs[i + 1 :]]
             pytest.fail(f"no row matching {expected} found")
 
         for row in expected_rows:
